/*
 *   A german multi-player version of Tetris for Microsoft Windows
 *   Compatible with the xtserv from xtris
 *
 *   Copyright (C) 2008 Matthias Ernst <m.tsnre@gmx.de>
 *
 *   Started: 12 Jul 2008
 *   Version: 1.0
 *
 *  Needed librarys to link: user32.lib gdi32.lib ws2_32.lib shell32.lib
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation. See the file COPYING for details.
 *
 */

//

//TODO: Chat-Nachrichten anzeigen/schreiben
//TODO: Bot-Hinzufuegen als Graphischen Button (nicht nur durch Taste 'B')
//Mgliche Features: Namen ndern; Tempo/Modus erzwingen; alle Bots automatisch killen; Variabler Speed-Modus; Lusitge Gewinn-Animation

#define _WIN32_WINNT 0x0500 //Windows 2000 oder hoeher
#define  WINVER      0x0500 //Windows 2000 oder hoeher
#define _CRT_SECURE_NO_DEPRECATE //Warnungen unterdruecken
#define BEZEICHNUNG "wtris v1.0"

#include <stdio.h> //Fuer die Konsolenausgabe
#include <stdlib.h> //Fuer Random
#include <time.h> //Zeit (Fuer Random-Initialisierung)
#include <winsock2.h> //Netzwerk-Sockets
#include <windows.h> //Generelle Windows-Funktionen
#include <shellapi.h> //Fuer ShellExecute (um Bot-Prozesse zu starten)

#include "resource.h" //Zum Zugriff auf die Eingabe-Dialoge der RC-Datei

#define PORT 19503
#define PROT_VERS_1 1
#define PROT_VERS_2 0
#define PROT_VERS_3 2
#define OP_NICK 1 //Namen senden
#define OP_PLAY 2 //(neues) Spiel starten
#define OP_FALL 3 //abgebaute Linien (stehen im Parameter), hoehere Linien fallen ^^
#define OP_DRAW 4 //Zu malende Steine (stehen im Parameter)
#define OP_LOST 5 //ok, ich habe verloren...
#define OP_GONE 6
#define OP_CLEAR 7 //Eigenes Feld leeren
#define OP_NEW 8
#define OP_LINES 9 //Anzahl der abgebauten Linien, die ein Gegner bekommen soll
#define OP_GROW 10 //Eigenes Feld komplett um eine Reihe nach oben schieben
#define OP_MODE 11 //0=normal, 1=fun, other=unknown
#define OP_LEVEL 12 //Geschwindigkeit ndern
#define OP_BOT 13 //Sagen, dass man ein Bot ist
#define OP_KILL 14 //Bot killen
#define OP_PAUSE 15
#define OP_CONT 16 //Spiel fortsetzen (Pause beenden)
#define OP_VERSION 17 //Protokoll-Version mitteilen/abgleichen
#define OP_BADVERS 18
#define OP_MSG 19 //Chat-Nachricht senden
#define OP_YOUARE 20
#define OP_LINESTO 21
#define OP_WON 22
#define OP_ZERO 23 //Spielstaende zuruecksetzen

#define STONE_START_X 2
#define STONE_START_Y -2

#define STONE_SIZE 16 //Groesse eines Steines in Pixel (mindestens 3)
#define WINDOW_HEIGHT STONE_SIZE*20+32+14 //Hhe des Main-Window
#define BUTTON_HEIGHT_FROM_TEXT 3 //Hoehe der Buttons ueber den Textinhalt hinaus
#define SPACE_BETWEEN_BUTTONS 4 //Platz zwischen den einzelnen Knpfen

#define TIMER_ID 34678 //willkuerlich
#define BG_COLOR 0xd8e9ec //Hintergrundfarbe des Fensters
#define SPIELFELD_BG_COLOR 0xb2b2b2 //Hintergrundfarbe des Spielfeldes

#define LOSE_ANIMATION_SPEED 50 //Lose-Animation: 50ms zwischen jeder Zeile ^^
#define LOSE_TIMER_ID 34679 //willkuerlich

#define MODUS_NORMAL 0
#define MODUS_FUN 1
#define MODUS_SPEED 2


enum STEIN{
	I_STONE=0,
	O_STONE=1,
	T_STONE=2,
	L_STONE=3,
	J_STONE=4,
	S_STONE=5,
	Z_STONE=6
};

enum DREHUNG{
	NORMAL=0,
	LEFT=1,
	AUFDEMKOPF=2,
	RIGHT=3
};
DREHUNG operator++(DREHUNG& dreh, int){ //Drehung gegen den Uhrzeigersinn
	DREHUNG tmp = dreh;
	if (dreh>=3)
		dreh=NORMAL;
	else
		dreh=(DREHUNG)(dreh+1);
	return tmp;
}
DREHUNG operator--(DREHUNG& dreh, int){ //Drehung im Uhrzeigersinn
	DREHUNG tmp = dreh;
	if (dreh<=0)
		dreh=RIGHT;
	else
		dreh=(DREHUNG)(dreh-1);
	return tmp;
}

struct fallingStone{
	STEIN steinArt;
	int x;
	int y;
	DREHUNG r;
};

struct Gegenspieler {
	int id;
	char name[16];
	int gewonnen;
	int linien;
	HDC Zeichenobjekt;
	HBITMAP TetrisBild;
	int loseAnimationStatus;
	bool loseAnimationActive;
};

struct Buttons {
	bool MousePressedItDown; //OnClick hat stattgefunden, Maus wurde noch nicht losgelassen
	bool DisplayItAsPressedDown; //MousePressedItDown ist erfuellt UND die Maus ist gerade ueber dem Knopf
	RECT Coords; //Position des Knopfes im Fenster
	char Caption[7]; //Text des Knopfes
	void (*pressedAction)(int param); //Pointer auf Funktion, die beim Druecken ausgefuehrt werden soll
};

char Spielfeld[10][20]; //Spielfeld beim Client ^^
fallingStone dieserStein; //Aktueller Stein
fallingStone naechsterStein; //Naechster Stein
int level=5; //aktuelles level
bool spielLaeuft=false; //Spieler spielt gerade
bool pause=false;
bool spielPausiert=false;
int meineSpielerID=0;
char meinSpielerName[16];
bool fun=false; //fun-Modus aktiv?
bool speed_mode=false; //speed-Modus aktiv?
int addlines=0; //Linien, die es als Strafe gibt...
int addfun=0; //fun-Bloecke, die es als Strafe gibt...

int gewonnen=0; //Anzahl der gewonnenen Spiele
int lines=0;  //Anzahl der abgebauten Linien
int loseAnimationStatus=-1;
bool loseAnimationActive=false;

int server_prot_revision=-1; //Was kann die Server-Protokollversion?

HDC Zeichenobjekt;
HBITMAP TetrisBild; //Hier wird die Ausgabe zwischengespeichert.
char* server; //IP-Addresse des Servers
SOCKET ClientSocket;
bool shiftIsPressed=false; //true, wenn Shift-Taste gerade gedrueckt wird

bool servereingabeStart;
bool namenseingabeStart;

RECT nextStone_Text = {6,6,0,0};
int SpielerStatusHeight=0; //Hoehe des Textes mit Spieler-Status-Infos
int ButtonTextHeight=0; //Hoehe des Textes fuer Buttons ^^

int anzahlGegner=0; //Anzahl der Gegner
Gegenspieler* gegner=NULL; //Daten der Gegner

Buttons knoepfe[13]; //Buttons zur Usersteuerung ^^
void ButtonStartPressed_Action(int param);
void ButtonPausePressed_Action(int param);
void ButtonSpeedPressed_Action(int param);
void ButtonModusPressed_Action(int param);
void ButtonResetPressed_Action(int param);

void toggleMode();

int Steine[7][4][8] = { //[SteinNr], [Drehung], [[X][Y]-Paare]

    /* ----
	   ++++
	   ----
	   ---- */
 {
    {  1, 2, 2, 2, 3, 2, 4, 2 },
    {  3, 0, 3, 1, 3, 2, 3, 3 },
    {  1, 2, 2, 2, 3, 2, 4, 2 },
    {  3, 0, 3, 1, 3, 2, 3, 3 }
 },
    /* ----
	   -++-
	   -++- 
	   ---- */
 {
    {  2, 3, 3, 3, 2, 2, 3, 2 },
    {  2, 3, 3, 3, 2, 2, 3, 2 },
    {  2, 3, 3, 3, 2, 2, 3, 2 },
    {  2, 3, 3, 3, 2, 2, 3, 2 }
 },

    /* ----
	   -+++
	   --+- 
	   ---- */
 {
    {  2, 2, 3, 2, 4, 2, 3, 3 },
    {  3, 2, 3, 3, 3, 4, 4, 3 },
    {  3, 2, 2, 3, 3, 3, 4, 3 },
    {  2, 3, 3, 2, 3, 3, 3, 4 }
 },

    /* ----
	   -+++
	   -+-- 
	   ---- */
 {
    {  2, 2, 3, 2, 4, 2, 2, 3 },
    {  3, 2, 3, 3, 3, 4, 4, 4 },
    {  2, 3, 3, 3, 4, 2, 4, 3 },
    {  2, 2, 3, 2, 3, 3, 3, 4 }
 },

    /* ----
	   -+++
	   ---+ 
	   ---- */
 {
    {  2, 2, 3, 2, 4, 2, 4, 3 },
    {  3, 2, 3, 3, 3, 4, 4, 2 },
    {  2, 2, 2, 3, 3, 3, 4, 3 },
    {  2, 4, 3, 2, 3, 3, 3, 4 }
 },

    /* ----
	   --++
	   -++- 
	   ---- */
 {
    {  2, 3, 3, 2, 3, 3, 4, 2 },
    {  3, 2, 3, 3, 4, 3, 4, 4 },
    {  2, 3, 3, 2, 3, 3, 4, 2 },
    {  3, 2, 3, 3, 4, 3, 4, 4 }
 },

    /* ----
	   -++-
	   --++ 
	   ---- */
 {
    {  2, 2, 3, 2, 3, 3, 4, 3 },
    {  3, 3, 3, 4, 4, 2, 4, 3 },
    {  2, 2, 3, 2, 3, 3, 4, 3 },
    {  3, 3, 3, 4, 4, 2, 4, 3 }
 }

};

int Steinfarben[9][3] = {
	{0x0000fa, 0x0000bc, 0x00007d}, //rot
	{0x00f000, 0x00a800, 0x007000}, //gruen
	{0xf08040, 0xd86000, 0x7c4000}, //blau
	{0x00b0d0, 0x008096, 0x006070}, //braun
	{0xb0b000, 0x808000, 0x606000}, //tuerkies
	{0xfa00fa, 0xbc00bc, 0x7d007d}, //pink
	{0xfa00bc, 0xbc007d, 0x7d001f}, //lila
	{SPIELFELD_BG_COLOR, SPIELFELD_BG_COLOR, SPIELFELD_BG_COLOR}, //grau (leerer Stein)
	{0xffffff, 0xb2b2b2, 0x808080}  //grau (fun-Block)
};

int levelSpeeds[9] = {1500, 1200, 900, 700, 500, 350, 200, 120, 80}; //Intervalle frs Runterfallen in Millisekunden

void cleanSpielfeld();
bool stoneMatch(fallingStone stone, bool* obenOwerflow=NULL); //Passt der Stein an die neue Stelle?
void placeStone(fallingStone stone); //Stein permanent im Level platzieren
bool moveStoneDown(fallingStone* stone, int* redraw=NULL); //Stein um eins nach unten bewegen (true bei Erfolg), bei false: neuen Stein einsetzen; redraw: x-y-Array(max. 2x8) mit neuzumalenden Positionen
void moveStoneToBottom(fallingStone* stone, int* redraw=NULL); //Stein ganz nach unten schmeissen und dort festsetzen; redraw: x-y-Array(max. 2x8) mit neuzumalenden Positionen
void rotateStone(fallingStone* stone, int* redraw=NULL); //Dreht den aktuellen Stein (wenn moeglich); redraw: x-y-Array(max. 2x8) mit neuzumalenden Positionen
void moveStoneRL(fallingStone* stone, bool right, int* redraw=NULL); //Bewegt den Stein um 1 nach links oder rechts; redraw: x-y-Array(max. 2x8) mit neuzumalenden Positionen
void moveStoneMaxRL(fallingStone* stone, bool right, int* redraw=NULL); //Bewegt den Stein so weit wie moeglich nach links oder rechts; redraw: x-y-Array(max. 2x8) mit neuzumalenden Positionen
bool createNewStone(fallingStone* stone, fallingStone* nextStone, int* redraw=NULL); //stone<-nextStone, nextStone<-Random, false->Niederlage!; Neuzumalen: Der komplette Stein (bei true, sonst ist GameOver)
void clearLine(int line); //Lsche eine Linie aus dem Spielfeld
int getNextCompleteLine(); //Gebe naechste vollstaendigie Zeile zurueck (-1, wenn keine existiert)
bool growUp(); //Schiebe alle Zeilen um eins nach oben, false, wenn es oben ueberlauft
void fillBottomLine(int* redraw=NULL); //Unterste Linie mit zufaelligen Steinen fuellen

void redrawLocalAndSendToServer(HWND hWnd, HDC Zeichenobjekt, int* redraw, SOCKET ClientSocket, bool neuerStein = false);
void ichHabeVerloren(SOCKET ClientSocket, HWND hWnd);
void legeNeuenSteinAn(SOCKET ClientSocket, HWND hWnd, int* redraw, HDC Zeichenobjekt);
bool addLines(HDC Zeichenobjekt, SOCKET ClientSocket, HWND hWnd);

bool SendBytes(SOCKET ClientSocket, void* buffer, int laenge); //Server-Kommunikation
void redrawForClearing(fallingStone* stone, int* redraw, int* akt);
void redrawForDrawing(fallingStone* stone, int* redraw, int* akt);

void createFunModeSides(int* redraw);
void createFunModeMiddle(int* redraw);

int calculateWindowWidth(int anzahlGegner);
bool changeWindowWidth(HWND hWnd, int newWidth);

int getInternGegnerNumber(int id);

//Callback-Fkt. fuer Eingabedialoge
INT_PTR CALLBACK EinagbeMessageHandler(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
//Callback-Fkt. fuer MainWindow (Spielschleife)
LRESULT CALLBACK MainDialogNachrichtenbehandlung(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

void DrawBlock(HDC Zeichenobjekt, int x, int y, int blockArt, int pixelX=1, int pixelY=1); //Malt einen Tetris-Block

void startServerAtLocalhost();


bool SendBytes(SOCKET ClientSocket, void* buffer, int laenge){
	if (laenge > 255){
		printf("Warnung: Das empfangene Kommando ist zu gross (%d Bytes). Die Daten wurden nicht uebermittelt.\n",laenge);
		return false;
	}
	int laenge_gedreht=0;
	((char*)(&laenge_gedreht))[0] = 0;
	((char*)(&laenge_gedreht))[1] = 0;
	((char*)(&laenge_gedreht))[2] = 0;
	((char*)(&laenge_gedreht))[3] = ((char*)(&laenge))[0];
	if (send(ClientSocket,(char*)&laenge_gedreht,4,0)==4)
		if (send(ClientSocket,(char*)buffer,laenge,0)==laenge)
			return true;
	return false;
}

void cleanSpielfeld(){
	for (int i=0;i<10;i++)
		for (int j=0;j<20;j++)
			Spielfeld[i][j]=7;
}

bool stoneMatch(fallingStone stone,bool* obenOwerflow){
	if (obenOwerflow!=NULL)
		*obenOwerflow=false;
	for (int i=0;i<4;i++){
		int x = Steine[stone.steinArt][stone.r][2*i]+stone.x;
		int y = Steine[stone.steinArt][stone.r][2*i+1]+stone.y;
		if (0>x || x>=10 || 0>y || y>=20){
			if (0>y)
				if (obenOwerflow!=NULL)
					*obenOwerflow=true;
			return false;
		}
		if (Spielfeld[x][y]!=7)
			return false;
	}
	return true;
}

void placeStone(fallingStone stone){
	for (int i=0;i<4;i++){
		int x = Steine[stone.steinArt][stone.r][2*i]+stone.x;
		int y = Steine[stone.steinArt][stone.r][2*i+1]+stone.y;
		/*if (0>x || x>=10 || 0>y || y>=20)
			continue; //Fehler
		if (Spielfeld[x][y]!=7)
			continue; //Fehler*/
		//Stein-Anteil platzieren
		Spielfeld[x][y] = (char)stone.steinArt;
	}
	return;
}

void redrawForClearing(fallingStone* stone, int* redraw, int* akt){
	if (redraw==NULL)
		return;
	//Alle bisherigen Orte merken (potetiell zu lschende)
	for (int i=0;i<4;i++){
		redraw[3*(*akt)]=stone->x+Steine[stone->steinArt][stone->r][2*i];
		redraw[3*(*akt)+1]=stone->y+Steine[stone->steinArt][stone->r][2*i+1];
		redraw[3*(*akt)+2]=7; //Hier loeschen
		if (redraw[3*(*akt)]>=0 && redraw[3*(*akt)+1]>=0)
			(*akt)++;
	}
	redraw[*akt*3]=-1;
}
void redrawForDrawing(fallingStone* stone, int* redraw, int* akt){
	if (redraw==NULL)
		return;
	int clearFeldMax=*akt;
	//Neue Position neu zeichnen ^^
			
	for (int i=0;i<4;i++){
		int x = Steine[stone->steinArt][stone->r][2*i]+stone->x;
		int y = Steine[stone->steinArt][stone->r][2*i+1]+stone->y;
		bool neuMalen=true;
		for (int j=0;j<clearFeldMax;j++){
			if (redraw[3*j]==x && redraw[3*j+1]==y){
				//Nicht neu zeichnen
				neuMalen=false;
				//Stein j muss auch nicht geloescht werden.
				(*akt)--;
				for (int k=j;k<*akt;k++){
					redraw[3*k] = redraw[3*k+3];
					redraw[3*k+1] = redraw[3*k+4];
					redraw[3*k+2] = redraw[3*k+5];
				}
				break;
			}
		}
		if (neuMalen){
			//Stein hinzufuegen
			redraw[*akt*3]=x;
			redraw[*akt*3+1]=y;
			redraw[*akt*3+2]=stone->steinArt;
			(*akt)++;
		}
	}
	redraw[*akt*3]=-1;
}

bool moveStoneDown(fallingStone* stone, int* redraw){
	int akt=0;
	redrawForClearing(stone,redraw,&akt);

	stone->y++;
	if (stoneMatch(*stone)){
		redrawForDrawing(stone,redraw,&akt);

		return true;
	}
	stone->y--;
	placeStone(*stone);
	if (redraw!=NULL)
		redraw[0]=-1;
	return false;
}
void moveStoneToBottom(fallingStone* stone, int* redraw){
	int akt=0;
	redrawForClearing(stone,redraw,&akt);

	do{
		stone->y++;
	}while(stoneMatch(*stone));
	stone->y--;

	redrawForDrawing(stone,redraw,&akt);
	placeStone(*stone);

}

bool createNewStone(fallingStone* stone, fallingStone* nextStone, int* redraw){
	*stone = *nextStone; //Naechsten Stein nehmen
	stone->r=NORMAL;
	stone->x=STONE_START_X;
	stone->y=STONE_START_Y;
	nextStone->steinArt=(STEIN)(rand()%7); //Zufaelliger naechster Stein
	if (stoneMatch(*stone)){
		int akt=0;
		redrawForDrawing(stone,redraw,&akt);
		return true;
	}
	if (redraw!=NULL)
		redraw[0]=-1;
	return false;
}

void rotateStone(fallingStone* stone, int* redraw){
	int akt=0;
	redrawForClearing(stone,redraw,&akt);

	stone->r++;
	bool obenOverflow=false;
	int goneDown=0;
	do{
		if (stoneMatch(*stone,&obenOverflow)){

			redrawForDrawing(stone,redraw,&akt);
			return;
		}
		if (obenOverflow){
			stone->y++; //Wenn er nicht passt: vllt. nach unten schieben?
			goneDown++;
		}
	}while(obenOverflow);
	//Er passt nicht, also rckgngig machen ^^
	stone->y -= goneDown;
	stone->r--;
	if (redraw!=NULL)
		redraw[0]=-1;
	return;
}

void moveStoneRL(fallingStone* stone, bool right, int* redraw){
	int akt=0;
	redrawForClearing(stone,redraw,&akt);
	stone->x+=(right?1:-1);
	if (stoneMatch(*stone)){
		redrawForDrawing(stone,redraw,&akt);
		return;
	}
	stone->x-=(right?1:-1);
	if (redraw!=NULL)
		redraw[0]=-1;
	return;
}

void moveStoneMaxRL(fallingStone* stone, bool right, int* redraw){
	int akt=0;
	redrawForClearing(stone,redraw,&akt);
	int moved=0;
	do{
		stone->x+=(right?1:-1);
		moved++;
	}while (stoneMatch(*stone));
	stone->x-=(right?1:-1);
	moved--;
	if (moved>0){
		redrawForDrawing(stone,redraw,&akt);
		return;
	}
	if (redraw!=NULL)
		redraw[0]=-1;
	return;
}

void clearLine(int line){
	for (int i=line; i>0; i--)
		for (int x=0;x<10;x++)
			Spielfeld[x][i] = Spielfeld[x][i-1]; //Hoehere Linie nachruecken
	for (int x=0;x<10;x++)
		Spielfeld[x][0] = 7; //Oberste Linie leeren
}

int getNextCompleteLine(){
	for (int y=0;y<20;y++){
		bool lueckeGefunden=false;
		for (int x=0;x<10;x++){
			if (Spielfeld[x][y]==7)
				lueckeGefunden=true;
		}
		if (!lueckeGefunden)
			return y;
	}
	return -1;
}

bool growUp(){
	for (int x=0;x<10;x++) //Oberste Linie pruefen
		if (Spielfeld[x][0] != 7)
			return false;
	for (int i=0; i<19; i++)
		for (int x=0;x<10;x++)
			Spielfeld[x][i] = Spielfeld[x][i+1]; //Tiefere Linie hochholen
	for (int x=0;x<10;x++)
		Spielfeld[x][19] = 7; //Unterste Linie leeren
	return true;
}

void fillBottomLine(int* redraw){
	int menge=0;
	if (redraw!=NULL){
		*redraw=-1;
	}
	for (int i=0;i<10;i++)
		if (Spielfeld[i][19]!=7)
			return;
	for (int i=0;i<10;i++){
		if (rand()%7 < 3){ //3/7-Chance auf Stein
			Spielfeld[i][19] = (char)(rand()%7); //Zufaelligen Stein nehmen
			redraw[3*menge]=i;
			redraw[3*menge+1]=19;
			redraw[3*menge+2]=Spielfeld[i][19];
			menge++;
		}
	}
	if (menge==10){
		//Es wurde ueberall aufgefuellt, also muss ein Stein wieder geloescht werden
		int stelle = rand()%10;
		Spielfeld[stelle][19] = 7;
		redraw[3*stelle+2] = 7; //Redraw-Feld auf 7 setzen, um umstaendlich rueckvberschieben zu vermeiden
	}
	redraw[3*menge]=-1;
}

void DrawBlock(HDC Zeichenobjekt, int x, int y, int blockArt, int pixelX, int pixelY){
	//Voraussetzung: Es wurde einmal folgendes aufgefhrt:
	//SelectObject(Zeichenobjekt,GetStockObject(DC_PEN));
	//SelectObject(Zeichenobjekt,GetStockObject(DC_BRUSH));

	//x- und y-Verschiebung merken
	const int xAdd = x*STONE_SIZE+pixelX;
	const int yAdd = y*STONE_SIZE+pixelY;

	//Farbe 1 waehlen
	SetDCPenColor(Zeichenobjekt,Steinfarben[blockArt][0]);
	//malen
	MoveToEx(Zeichenobjekt, STONE_SIZE-1 + xAdd, yAdd, 0);
	LineTo(Zeichenobjekt, xAdd, yAdd);
	LineTo(Zeichenobjekt, xAdd, STONE_SIZE-1 + yAdd);

	//Farbe 2 waehlen (PEN & BRUSH)
	SetDCBrushColor(Zeichenobjekt,Steinfarben[blockArt][1]);
	SetDCPenColor(Zeichenobjekt,Steinfarben[blockArt][1]);
	//malen
	Rectangle(Zeichenobjekt,1 + xAdd,1 + yAdd,STONE_SIZE-1 + xAdd,STONE_SIZE-1 + yAdd);
	
	//Farbe 3 waehlen
	SetDCPenColor(Zeichenobjekt,Steinfarben[blockArt][2]);
	//malen
	MoveToEx(Zeichenobjekt,xAdd, STONE_SIZE-1 + yAdd,0);
	LineTo(Zeichenobjekt, STONE_SIZE-1 + xAdd, STONE_SIZE-1 + yAdd);
	LineTo(Zeichenobjekt, STONE_SIZE-1 + xAdd, yAdd);
}

void redrawLocalAndSendToServer(HWND hWnd, HDC Zeichenobjekt, int* redraw, SOCKET ClientSocket, bool neuerStein){
	char buffer[255];
	buffer[0]=0;
	buffer[1]=OP_DRAW;
	int i=0;
	while (redraw[i]>=0){
		buffer[i+2]=redraw[i];
		i++;
	}
	if (i>0){
		SendBytes(ClientSocket,buffer,i+2);
	}

	i=0;
	while (redraw[i]>=0){
		DrawBlock(Zeichenobjekt, redraw[i], redraw[i+1], redraw[i+2]);
		i+=3;
	}
	if (neuerStein){
		InvalidateRect(hWnd,NULL,FALSE);
	}else if (i>0){
		RECT SpielerBereich = {nextStone_Text.right+nextStone_Text.left, 6, nextStone_Text.right+nextStone_Text.left+10*STONE_SIZE+2, 6+20*STONE_SIZE+2};
		InvalidateRect(hWnd,&SpielerBereich,FALSE);
	}
}

void ichHabeVerloren(SOCKET ClientSocket, HWND hWnd){
	if (spielLaeuft){
		char buffer[2];
		buffer[0]=0;
		buffer[1]=OP_LOST;
		SendBytes(ClientSocket,buffer,2);
		spielLaeuft = false;
		KillTimer(hWnd, TIMER_ID);
		printf("Ich habe verloren... -_-\n");
		//"Neuen Stein"-Feld leeren
		InvalidateRect(hWnd,NULL,FALSE);
		//Lose-Animation anzeigen
		loseAnimationStatus=-1;
		SetTimer(hWnd, LOSE_TIMER_ID, LOSE_ANIMATION_SPEED, NULL);
		loseAnimationActive=true;
	}
}

void legeNeuenSteinAn(SOCKET ClientSocket, HWND hWnd, int* redraw, HDC Zeichenobjekt){
	int abbaubar = 0;
	int zeile=-1;

	char buffer[6];
	buffer[0]=0;
	buffer[1]=OP_FALL;

	while ((zeile=getNextCompleteLine())>=0){
		clearLine(zeile); //Abbauen (nur im Speicher, nicht graphisch)
		BitBlt(Zeichenobjekt,1,STONE_SIZE+1,10*STONE_SIZE,STONE_SIZE*zeile,Zeichenobjekt,1,1,SRCCOPY);
		SetDCBrushColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
		SetDCPenColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
		Rectangle(Zeichenobjekt, 1, 1, 10*STONE_SIZE+1, STONE_SIZE+1);

		buffer[abbaubar+2]=zeile;
		abbaubar++;
	}
	if (abbaubar>0){ //Malen & An den Server schicken
		RECT SpielerBereich = {nextStone_Text.right+nextStone_Text.left, 6, nextStone_Text.right+nextStone_Text.left+10*STONE_SIZE+2, 6+20*STONE_SIZE+2};
		InvalidateRect(hWnd,&SpielerBereich,FALSE);
		lines+=abbaubar;
		SendBytes(ClientSocket,buffer,abbaubar+2);
		if (fun || abbaubar>1){
			buffer[1]=OP_LINES;
			buffer[2]=(char)abbaubar;
			SendBytes(ClientSocket,buffer,3);
		}
	}
	
	//Neuer Stein: Timer resetten (killen & neu anlegen)
	KillTimer(hWnd, TIMER_ID);
	SetTimer(hWnd, TIMER_ID, levelSpeeds[level-1], NULL);
	//ggf. Straflinien hinzufuegen
	if (!addLines(Zeichenobjekt,ClientSocket,hWnd)){
		ichHabeVerloren(ClientSocket,hWnd);
		return;
	}
	//neuen Stein anlegen
	if (createNewStone(&dieserStein,&naechsterStein,redraw)){
		redrawLocalAndSendToServer(hWnd, Zeichenobjekt, redraw, ClientSocket, true);
	}else{
		//VERLOREN!!!
		ichHabeVerloren(ClientSocket,hWnd);
	}
}

bool addLines(HDC Zeichenobjekt, SOCKET ClientSocket, HWND hWnd){
	char buffer[2];
	while(addlines>0){							
		if (!growUp()){
			ichHabeVerloren(ClientSocket,hWnd);
			addlines=0;
			return false;
		}else{
			//grow-Up: Graphisch und an den Server senden ^^
			//malen
			BitBlt(Zeichenobjekt,1,1,10*STONE_SIZE,STONE_SIZE*19,Zeichenobjekt,1,STONE_SIZE+1,SRCCOPY);
			SetDCBrushColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
			SetDCPenColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
			Rectangle(Zeichenobjekt, 1, 19*STONE_SIZE+1, 10*STONE_SIZE+1, 20*STONE_SIZE+1);
			//Fenster-Update
			RECT SpielerBereich = {nextStone_Text.right+nextStone_Text.left, 6, nextStone_Text.right+nextStone_Text.left+10*STONE_SIZE+2, 6+20*STONE_SIZE+2};
			InvalidateRect(hWnd,&SpielerBereich,FALSE);
			//Zum Server
			buffer[1] = OP_GROW;
			SendBytes(ClientSocket,buffer,2);
			//Unterste Zeile zufllig fllen...
			int redraw[31];
			fillBottomLine(redraw);
			redrawLocalAndSendToServer(hWnd, Zeichenobjekt,redraw,ClientSocket);
			addlines--;
		}
	}
	while(addfun>0){
		int y=0;
		int x=0;
		for (;y<19;y++){
			for (x=0;x<9;x++)
				if (Spielfeld[x][y]!=7)
					break;
			if (Spielfeld[x][y]!=7)
				break;
		}
		y--;
		if (y<0)
			return false;
		Spielfeld[x][y]=8; //fun-Block platzieren ^^
		int redraw[4];
		redraw[0]=x;
		redraw[1]=y;
		redraw[2]=8;
		redraw[3]=-1;
		redrawLocalAndSendToServer(hWnd, Zeichenobjekt,redraw,ClientSocket);
		addfun--;
	}
	return true;
}

void createFunModeSides(int* redraw){
	for (int i=10; i<20; i++) {
		Spielfeld[0][i] = rand() % 7;
		Spielfeld[9][i] = rand() % 7;
		redraw[6*(i-10)]=0;
		redraw[6*(i-10)+1]=i;
		redraw[6*(i-10)+2]=Spielfeld[0][i];
		redraw[6*(i-10)+3]=9;
		redraw[6*(i-10)+4]=i;
		redraw[6*(i-10)+5]=Spielfeld[9][i];
	}
	redraw[3*20]=-1;
}

void createFunModeMiddle(int* redraw){
	int anzahlSteine=0;
	for (int i=0; i<24; i++){
		int x=rand()%10;
		int y=15+rand()%5;
		int stein = rand()%7;
		Spielfeld[x][y] = (char)stein;
		for (int j=0;j<anzahlSteine;j++){
			if (redraw[3*j]==x && redraw[3*j+1]==y){
				//Der Stein sollte nur einmal neu gezeichnet werden.
				anzahlSteine--;
				for (int k=j;k<anzahlSteine;k++){
					redraw[3*k] = redraw[3*k+3];
					redraw[3*k+1] = redraw[3*k+4];
					redraw[3*k+2] = redraw[3*k+5];
				}
				break; //Das Feld kam, wenn es schon existierte, nur 1x vor
			}
		}
		redraw[3*anzahlSteine] = x;
		redraw[3*anzahlSteine+1] = y;
		redraw[3*anzahlSteine+2] = stein;
		anzahlSteine++;
	}
	redraw[3*anzahlSteine] = -1;
}

int calculateWindowWidth(int anzahlGegner){
	int breite = STONE_SIZE*10+20+nextStone_Text.right;
	if (breite < STONE_SIZE*14+28)
		breite = STONE_SIZE*14+28;
	//Jetzt fr Gegner verbreitern
	breite+=anzahlGegner*(STONE_SIZE*10+6);
	return breite;
}

bool changeWindowWidth(HWND hWnd, int newWidth){
	WINDOWPLACEMENT oldPos;
	oldPos.length=sizeof(WINDOWPLACEMENT);
	BOOL erfolg = GetWindowPlacement(hWnd, &oldPos);
	if (erfolg==FALSE)
		return false;

	erfolg = MoveWindow(hWnd,oldPos.rcNormalPosition.left,oldPos.rcNormalPosition.top,newWidth,WINDOW_HEIGHT+6+SpielerStatusHeight,TRUE);
	return erfolg==TRUE?true:false;

}

int getInternGegnerNumber(int id){
	int position = 0;
	while(position < anzahlGegner){
		if (gegner[position].id==id)
			break;
		position++;
	}
	return position;
}

void startServerAtLocalhost(){
	printf("Versuche, einen xtris-Server auf localhost zu starten.\n");
	ShellExecute(NULL, NULL, "xtserv.exe", "-once", NULL, SW_HIDE);
	Sleep(1000);
}

//Knopf-Behandlungen
void ButtonStartPressed_Action(int param){
	char buffer[2];
	buffer[0]=0;
	buffer[1]=OP_PLAY;
	SendBytes(ClientSocket,buffer,2);
}

void ButtonPausePressed_Action(int param){
	char buffer[3];
	buffer[0]=0;
	buffer[1]=spielPausiert?OP_CONT:OP_PAUSE;
	SendBytes(ClientSocket,buffer,2);
}

void ButtonSpeedPressed_Action(int param){
	char buffer[3];
	buffer[0]=0;
	buffer[1]=OP_LEVEL;
	buffer[2]=(char)(param-1);
	SendBytes(ClientSocket,buffer,3);
}

void ButtonModusPressed_Action(int param){
	toggleMode();
}

void ButtonResetPressed_Action(int param){
	char buffer[2];
	buffer[0]=0;
	buffer[1]=OP_ZERO;
	SendBytes(ClientSocket,buffer,2);
}

INT_PTR CALLBACK EinagbeMessageHandler(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam){
	switch (uMsg){
		case WM_CLOSE:
			DestroyWindow(hWnd);
			return true;
		case WM_DESTROY:
			EndDialog(hWnd,0);
			return true;
		case WM_COMMAND:
			if (LOWORD(wParam)==IDC_BUTTON1){
				//Text des Eingabefeldes ermitteln ^^
				char* text = new char[16];
				GetDlgItemTextA(hWnd,IDC_EDIT1,text,16);
				EndDialog(hWnd,(INT_PTR)text);
				return true;
			}
			return false;
		case WM_ACTIVATE:
			{
				HWND editFeld = GetDlgItem(hWnd, IDC_EDIT1);
				SetFocus(editFeld);
				if (namenseingabeStart){
					//alten Namen wenn vorhanden aus Datei laden
					FILE* datei = NULL;
					try{
						datei = fopen("wtris.conf","rb");
						if (datei!=NULL){
							int laenge=0;
							fread(&laenge,sizeof(int),1,datei);
							if (laenge > 0){
								char* name = new char[laenge+1];
								fread(name,1,laenge,datei);
								name[laenge]=0;
								SetDlgItemTextA(hWnd,IDC_EDIT1,name);
								SendMessage(editFeld,EM_SETSEL,0,-1);
								delete name;
							}
							fclose(datei);
						}
					}catch(...){}
					namenseingabeStart=false;
				}else if (servereingabeStart){
					//alte Server-IP laden
					FILE* datei = NULL;
					try{
						datei = fopen("wtris.conf","rb");
						if (datei!=NULL){
							int laenge=0;
							fread(&laenge,sizeof(int),1,datei);
							if (laenge >= 0){
								fseek(datei,laenge,SEEK_CUR);
								fread(&laenge,sizeof(int),1,datei);
								if (laenge > 0){
									char* name = new char[laenge+1];
									fread(name,1,laenge,datei);
									name[laenge]=0;
									SetDlgItemTextA(hWnd,IDC_EDIT1,name);
									SendMessage(editFeld,EM_SETSEL,0,-1);
									delete name;
								}
							}
							fclose(datei);
						}
					}catch(...){}
					servereingabeStart=false;
				}
			}
			return false;
	}
	return false;
}

//Nachrichtenbehandlung fuers MainWindow
LRESULT CALLBACK MainDialogNachrichtenbehandlung(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam){
	switch (uMsg){
		case WM_CLOSE:
			DestroyWindow(hWnd);
			return 0;
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;
		case WM_KEYDOWN:
			switch (wParam){
				case VK_SHIFT:
					shiftIsPressed=true;
					break;
				case VK_F2:
				case 'N':
				case 'S':
					{
						//'N', 'S' oder F2: Neues Spiel starten
						char buffer[2];
						buffer[0]=0;
						buffer[1]=OP_PLAY;
						SendBytes(ClientSocket,buffer,2);
						break;
					}
				case VK_F3:
				case 'P':
				case VK_PAUSE:
					{
						//F3 oder 'P' oder Unterbr.: Pause/Resume
						char buffer[3];
						buffer[0]=0;
						buffer[1]=spielPausiert?OP_CONT:OP_PAUSE;
						SendBytes(ClientSocket,buffer,2);
						break;
					}
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
					{
						//'1'-'9': Tempo setzen
						char buffer[3];
						buffer[0]=0;
						buffer[1]=OP_LEVEL;
						buffer[2]=(char)(wParam-'0');
						SendBytes(ClientSocket,buffer,3);
						break;
					}
				case 'F':
				case 'M':
					{
						//'M' oder 'F': fun/speed ein-/ausschalten
						toggleMode();
						break;
					}
				case 'C':
				case 'Z':
				case 'R':
					{
						//'C','Z','R': Statistiken zuruecksetzen
						char buffer[2];
						buffer[0]=0;
						buffer[1]=OP_ZERO;
						SendBytes(ClientSocket,buffer,2);
						break;
					}
				case 'B':
					{
						//'B': Bot-Prozess starten
						printf("Versuche, einen Bot-Prozess zu starten.\n");
						ShellExecute(NULL, NULL, "xtbot.exe", server, NULL, SW_HIDE);
					}
			}
			if (spielLaeuft){
			//if (((1<<30)&lParam)==0 && spielLaeuft){ //Taste vorher nicht gedrueckt
				int redraw[25];
				*redraw=-1;
				bool newStone=false;
				switch (wParam){
					case VK_LEFT:
						//Taste-links
						if (shiftIsPressed)
							moveStoneMaxRL(&dieserStein,false,redraw);
						else
							moveStoneRL(&dieserStein,false,redraw);
						break;
					case VK_RIGHT:
						//Taste-rechts
						if (shiftIsPressed)
							moveStoneMaxRL(&dieserStein,true,redraw);
						else
							moveStoneRL(&dieserStein,true,redraw);
						break;
					case VK_UP:
						//Taste-hoch
						rotateStone(&dieserStein,redraw);
						break;
					case VK_DOWN:
						//Taste-Unten
						if (!moveStoneDown(&dieserStein, redraw))
							newStone=true;
						break;
					case VK_SPACE:
						//Space
						moveStoneToBottom(&dieserStein, redraw);
						newStone=true;
						break;
				}
				//neu malen...
				redrawLocalAndSendToServer(hWnd, Zeichenobjekt, redraw, ClientSocket);

				if (newStone){
					//neuen Stein anlegen
					legeNeuenSteinAn(ClientSocket, hWnd, redraw, Zeichenobjekt);
				}
			}
			return 0;
		case WM_KEYUP:
			if (wParam==VK_SHIFT)
				shiftIsPressed=false;
			return 0;
		case WM_LBUTTONDOWN:
			{
				//Mausklick
				int x=LOWORD(lParam);
				int y=HIWORD(lParam);

				//Linken Fensterbereich abziehen
				x-=nextStone_Text.right+nextStone_Text.left-6;
				if (x>0){
					int spieler = x/(STONE_SIZE*10+8)-1;
					if (x%(STONE_SIZE*10+8)>=6){
						if (y>=6 && y<=STONE_SIZE*20+14+SpielerStatusHeight){
							if (spieler < anzahlGegner && spieler >= 0){
								//Spieler "spieler" wurde angeklickt.
								//Also versuche, ihn zu KILLEN! ^_^
								char buffer[3];
								buffer[0]=0;
								buffer[1]=OP_KILL;
								buffer[2]=gegner[spieler].id;
								SendBytes(ClientSocket,buffer,3);
								printf("Versuche, %s zu killen.\n",gegner[spieler].name);
								return 0;
							}

						}
					}
				}
				x=LOWORD(lParam);
				//Vllt. wurde auf einen Button geklickt? ^_^
				for (int i=0;i<13;i++){
					if (y >= knoepfe[i].Coords.top && y <= knoepfe[i].Coords.bottom && x >= knoepfe[i].Coords.left && x <= knoepfe[i].Coords.right){
						//Ja, es wurde auf einen Knopf geklickt ^^
						knoepfe[i].MousePressedItDown=true;
						knoepfe[i].DisplayItAsPressedDown=true; //Mit MouseMove beobachten!!!
						//Knopf neu zeichnen! ^^
						InvalidateRect(hWnd,&(knoepfe[i].Coords),FALSE);
					}
				}
				return 0;
			}
		case WM_MOUSEMOVE:
			{
				//Cursor fr Buttons beobachten: vllt. ist ja ein Button geklickt und dieser wird verlassen oder neu erreicht?
				int x=LOWORD(lParam);
				int y=HIWORD(lParam);

				for (int i=0; i<13; i++){
					if (knoepfe[i].MousePressedItDown){
						bool MausIstDaDrauf = (y >= knoepfe[i].Coords.top && y <= knoepfe[i].Coords.bottom && x >= knoepfe[i].Coords.left && x <= knoepfe[i].Coords.right);
						if (knoepfe[i].DisplayItAsPressedDown && !MausIstDaDrauf){
							knoepfe[i].DisplayItAsPressedDown = false;
							InvalidateRect(hWnd,&(knoepfe[i].Coords),FALSE);
						}else if (!knoepfe[i].DisplayItAsPressedDown && MausIstDaDrauf){
							knoepfe[i].DisplayItAsPressedDown = true;
							InvalidateRect(hWnd,&(knoepfe[i].Coords),FALSE);
						}
					}
				}
				return 0;
			}
		case WM_LBUTTONUP:
			{
				//Maus losgelassen
				int x=LOWORD(lParam);
				int y=HIWORD(lParam);

				//Vllt. ja ueber einem gedrueckten Knopf?
				for (int i=0; i<13;i++){
					if (knoepfe[i].MousePressedItDown){
						//Jeden gedrueckten Knopf als nicht mehr gedrueckt festlegen ^^
						knoepfe[i].MousePressedItDown=false;
						if (knoepfe[i].DisplayItAsPressedDown){
							knoepfe[i].DisplayItAsPressedDown=false;
							 knoepfe[i].pressedAction(i);
							InvalidateRect(hWnd,&(knoepfe[i].Coords),FALSE);
						}
					}
				}
				return 0;
			}
		case WM_TIMER:
			if (wParam==TIMER_ID && spielLaeuft){
				int redraw[25];
				*redraw=-1;
				bool newStone=false;
				//Stein nach unten fallen lassen, evtl. neuer Stein, evtl. verlieren
				if (!moveStoneDown(&dieserStein, redraw)){
					newStone=true;
				}
				//neu malen
				redrawLocalAndSendToServer(hWnd, Zeichenobjekt, redraw, ClientSocket);

				if (newStone){
					//neuen Stein anlegen
					legeNeuenSteinAn(ClientSocket, hWnd, redraw, Zeichenobjekt);
				}
				return 0;
			}
			if (wParam >= LOSE_TIMER_ID && wParam <= LOSE_TIMER_ID+(unsigned int)anzahlGegner){
				//Lose-Animation fuer Gegner (wParam-LOSE_TIMER_ID-1) bearbeiten
				//Dabei ist der Gegner Nr. -1 der Spieler.
				int spieler = (int)(wParam-(unsigned int)LOSE_TIMER_ID-1);
				HDC malobjekt = (spieler==-1?Zeichenobjekt:gegner[spieler].Zeichenobjekt);
				int* animationStatus = (spieler==-1?&loseAnimationStatus:&(gegner[spieler].loseAnimationStatus));
				bool* active = (spieler==-1?&loseAnimationActive:&(gegner[spieler].loseAnimationActive));

				int rechts = nextStone_Text.right+nextStone_Text.left;
				//TODO: BEREICH EVTL. NOCH BESSER WHLEN!!!
				RECT neuMalBereich={nextStone_Text.right+nextStone_Text.left-6+10*STONE_SIZE+14+(STONE_SIZE*10+6)*spieler-(spieler<0?1:0),6,0,STONE_SIZE*20+8};//{rechts+(spieler+1)*(STONE_SIZE*10+8),6,0,20*STONE_SIZE+8};
				neuMalBereich.right = neuMalBereich.left + 2 + STONE_SIZE*10;

				(*animationStatus)++;
				//Jetzt: Animation malen!!!
				if ((*animationStatus) < 20){
					//Zeile 19-(*animationStatus) mit fun-Blcken fllen
					int zeile = 19-(*animationStatus);
					for (int i=0;i<10;i++){
						DrawBlock(malobjekt,i,zeile,8);
					}
				}else if ((*animationStatus) < 40){
					//Zeile (*animationStatus)-20 leeren
					int zeile = (*animationStatus)-20;
					SetDCBrushColor(malobjekt, SPIELFELD_BG_COLOR);
					SetDCPenColor(malobjekt, SPIELFELD_BG_COLOR);
					Rectangle(malobjekt, 1, 1+zeile*STONE_SIZE, 10*STONE_SIZE+1, (zeile+1)*STONE_SIZE+1);

				}
				if ((*animationStatus)>=39){
					//Letzte Animation wurde gemalt ^^
					*animationStatus = -1;
					KillTimer(hWnd, wParam);
					*active=false;
				}
				InvalidateRect(hWnd,&neuMalBereich,FALSE);
				return 0;
			}
			break;
		case WM_PAINT:
			{
				RECT updateRegion;
				if (GetUpdateRect(hWnd, &updateRegion, FALSE)==0){
					return 0;
				}
				char statistikAusgabeString[60];
				PAINTSTRUCT ps;
				HDC fenster = BeginPaint(hWnd, &ps);
				SelectObject(fenster,GetStockObject(DC_PEN));
				SelectObject(fenster,GetStockObject(DC_BRUSH));
				SetBkColor(fenster,BG_COLOR);

				int rechts = nextStone_Text.right+nextStone_Text.left-6;
				int middle = (nextStone_Text.left+(nextStone_Text.right-nextStone_Text.left)/2);

				DrawTextA(fenster,"Nchster Stein:",-1,&nextStone_Text,DT_NOCLIP);
				SetDCBrushColor(fenster, 0xffffff);
				SetDCPenColor(fenster, 0x000000);
				
				//Rechteck um nchsten Stein ermitteln
				RECT nextStoneRect = {middle-2*STONE_SIZE-1,nextStone_Text.bottom,middle+2*STONE_SIZE+1,nextStone_Text.bottom+STONE_SIZE*2+2};
				//Nchsten Stein nur neu malen, wenn er im Invalidate-Rect liegt
				if (updateRegion.left <= nextStoneRect.right && updateRegion.right >= nextStoneRect.left && updateRegion.top <= nextStoneRect.bottom && updateRegion.bottom >= nextStoneRect.top)
				{
					//Rechteck um nchsten Stein malen
					Rectangle(fenster,nextStoneRect.left,nextStoneRect.top,nextStoneRect.right,nextStoneRect.bottom);
					if (spielLaeuft || pause){
						//Nchsten Stein malen
						for (int i=0;i<4;i++){
							DrawBlock(fenster,Steine[naechsterStein.steinArt][0][0+2*i]-1,
								Steine[naechsterStein.steinArt][0][1+2*i]-2,naechsterStein.steinArt,middle-2*STONE_SIZE,nextStone_Text.bottom+1);
						}
					}
				}

				//Speed & Modus anzeigen
				char* text = new char[24+13];
				memcpy(text,"Speed: 5\r\nModus: normal\r\nSpeed-Modus",24+13);
				RECT speedAndModeRect = {6,nextStone_Text.bottom+STONE_SIZE*2+8,0,0};
				DrawTextA(fenster,text,-1,&speedAndModeRect,DT_CALCRECT|DT_NOCLIP);
				text[7]=(char)(level+'0');
				if (!speed_mode){
					if (fun){
						text[17]='f';
						text[18]='u';
						text[19]='n';
						text[20]=0;
					}
					else
						text[23]=0;
				}else if (fun){
						text[17]='f';
						text[18]='u';
						text[19]='n';
						memcpy(text+20,text+23,14);
				}
					RECT biggerRect = speedAndModeRect;
					if (fun)
						text[20]=0;
					DrawTextA(fenster,text,-1,&speedAndModeRect,DT_CALCRECT|DT_NOCLIP);
					if (speed_mode && fun)
						text[20]='\r';
					biggerRect.left = speedAndModeRect.right;
					if (biggerRect.left != biggerRect.right){
						SetDCBrushColor(fenster, BG_COLOR);
						SetDCPenColor(fenster, BG_COLOR);
						Rectangle(fenster,biggerRect.left,biggerRect.top,biggerRect.right+1,biggerRect.bottom+1);
				}
				biggerRect = speedAndModeRect;
				if (!speed_mode){
					if (fun){
						text[20]='X';
					}
					else
						text[23]='\r';
				}
				DrawTextA(fenster,text,-1,&speedAndModeRect,DT_CALCRECT|DT_NOCLIP);
				if (!speed_mode){
					if (fun){
						text[20]=0;
					}
					else
						text[23]=0;
				}
				biggerRect.top = speedAndModeRect.bottom;
				if (biggerRect.top != biggerRect.bottom){
						SetDCBrushColor(fenster, BG_COLOR);
						SetDCPenColor(fenster, BG_COLOR);
						Rectangle(fenster,biggerRect.left,biggerRect.top,biggerRect.right+1,biggerRect.bottom+1);
				}

				DrawTextA(fenster,text,-1,&speedAndModeRect,DT_NOCLIP);
				delete text;

				//Knoepfe malen ^^
				for (int i=0; i<13; i++){
					if (updateRegion.left <= knoepfe[i].Coords.right && updateRegion.right >= knoepfe[i].Coords.left && updateRegion.top <= knoepfe[i].Coords.bottom && updateRegion.bottom >= knoepfe[i].Coords.top){
						//Knopf i neu malen ^^
						if (knoepfe[i].DisplayItAsPressedDown){
							SetDCBrushColor(fenster, 0x000000);
							SetBkColor(fenster,0x000000);
							SetDCPenColor(fenster, 0xffffff);
							SetTextColor(fenster,0xffffff);
						}else{
							SetDCBrushColor(fenster, 0xffffff);
							SetBkColor(fenster,0xffffff);
							SetDCPenColor(fenster, 0x000000);
							SetTextColor(fenster,0x000000);
						}
						Rectangle(fenster,knoepfe[i].Coords.left,knoepfe[i].Coords.top,knoepfe[i].Coords.right,knoepfe[i].Coords.bottom);
						RECT textRect = knoepfe[i].Coords;
						textRect.top += BUTTON_HEIGHT_FROM_TEXT;
						DrawTextA(fenster,knoepfe[i].Caption,-1,&textRect,DT_NOCLIP|DT_CALCRECT);
						int nachRechtsSchieben = (knoepfe[i].Coords.right - textRect.right)/2;
						textRect.left += nachRechtsSchieben;
						textRect.right += nachRechtsSchieben;
						DrawTextA(fenster,knoepfe[i].Caption,-1,&textRect,DT_NOCLIP);
					}
				}
				SetBkColor(fenster,BG_COLOR);
				SetTextColor(fenster,0x000000);

				//Spieler-Spielfeld malen
				BitBlt(fenster,rechts+6,6,10*STONE_SIZE+2,20*STONE_SIZE+2,Zeichenobjekt,0,0,SRCCOPY);
				//Spieler-Statistiken malen?
				RECT spielerStatBereich = {rechts+6,14+20*STONE_SIZE,rechts+10*STONE_SIZE+8,14+20*STONE_SIZE+SpielerStatusHeight};
				if (updateRegion.left <= spielerStatBereich.right && updateRegion.right >= spielerStatBereich.left && updateRegion.top <= spielerStatBereich.bottom && updateRegion.bottom >= spielerStatBereich.top){
					RECT statistik = {0,14+20*STONE_SIZE,0,0};

						sprintf(statistikAusgabeString,"%s\r\n%d %s\r\n%d %s",meinSpielerName,lines,lines==1?"Linie":"Linien",gewonnen,gewonnen==1?"Spiel":"Spiele");
						DrawTextA(fenster,statistikAusgabeString,-1,&statistik,DT_CALCRECT|DT_NOCLIP);
						int leftStart = (rechts+7+5*STONE_SIZE-statistik.right/2);
						statistik.left+=leftStart;
						statistik.right+=leftStart;
					
						if (statistik.bottom-statistik.top!=SpielerStatusHeight){
							printf("Warnung: Spielerinfos an Grundposition sind hoeher als erwartet.\n");
						}		

						SetDCBrushColor(fenster, BG_COLOR);
						SetDCPenColor(fenster, BG_COLOR);
						Rectangle(fenster,spielerStatBereich.left,spielerStatBereich.top,spielerStatBereich.right+1,spielerStatBereich.bottom+1);

						DrawTextA(fenster,statistikAusgabeString,-1,&statistik,DT_NOCLIP|DT_CENTER);
				}

				//Gegner-Spielfelder malen!
				int rechterRandDesSpielerSpielfeldes = rechts+10*STONE_SIZE+14;
				for (int i=0;i<anzahlGegner;i++){
					BitBlt(fenster,rechterRandDesSpielerSpielfeldes+(STONE_SIZE*10+6)*i,6,10*STONE_SIZE+2,20*STONE_SIZE+2,gegner[i].Zeichenobjekt,0,0,SRCCOPY);
					
					//Statistik neu malen???
					RECT statistikBereich = {rechterRandDesSpielerSpielfeldes+(STONE_SIZE*10+6)*i,14+20*STONE_SIZE,0,14+20*STONE_SIZE+SpielerStatusHeight};
					statistikBereich.right = statistikBereich.left+2+20*STONE_SIZE;
					if (updateRegion.left <= statistikBereich.right && updateRegion.right >= statistikBereich.left && updateRegion.top <= statistikBereich.bottom && updateRegion.bottom >= statistikBereich.top){

						//Spielerstatistik ausgeben
						RECT statistik = {0,14+20*STONE_SIZE,0,0};
						sprintf(statistikAusgabeString,"%s\r\n%d %s\r\n%d %s",gegner[i].name,gegner[i].linien,gegner[i].linien==1?"Linie":"Linien",gegner[i].gewonnen,gegner[i].gewonnen==1?"Spiel":"Spiele");
						DrawTextA(fenster,statistikAusgabeString,-1,&statistik,DT_CALCRECT|DT_NOCLIP);
						int leftStart = (rechterRandDesSpielerSpielfeldes+(STONE_SIZE*10+6)*i+5*STONE_SIZE+1-statistik.right/2);
						statistik.left+=leftStart;
						statistik.right+=leftStart;
					
						if (statistik.bottom-statistik.top!=SpielerStatusHeight){
							printf("Warnung: Spielerinfos an Position %d sind hoeher als erwartet.\n",i);
						}		

						SetDCBrushColor(fenster, BG_COLOR);
						SetDCPenColor(fenster, BG_COLOR);
						Rectangle(fenster,statistikBereich.left,statistikBereich.top,statistikBereich.right+1,statistikBereich.bottom+1);

						DrawTextA(fenster,statistikAusgabeString,-1,&statistik,DT_NOCLIP|DT_CENTER);
					}
				}
				EndPaint(hWnd, &ps);
				//return DefWindowProc(hWnd,uMsg,wParam,lParam);
				return 0;
			}

	}
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

int main(int argc, char* argv[])
{
	srand((unsigned int)time(NULL));
	HINSTANCE hInstance = GetModuleHandle(NULL);

	HWND MainWindow=0;

	servereingabeStart=false;
	namenseingabeStart=true;
	char* name = (char*)DialogBox(hInstance,MAKEINTRESOURCE(NAMENSEINGABE),NULL,EinagbeMessageHandler);
	if (name==NULL){
		return -1;
	}
	if (*name==0){
		delete name;
		return -1;
	}
	int namelen=(int)strlen(name);

	memcpy(meinSpielerName,name,namelen<16?namelen+1:15);
	meinSpielerName[15]=0;
	if (namelen >= 16)
		namelen=15;
	delete name;

	bool noServerEntered=false;
	servereingabeStart=true;
	namenseingabeStart=false;
	server = (char*)DialogBox(hInstance,MAKEINTRESOURCE(SERVERIP),NULL,EinagbeMessageHandler);
	if (server==NULL){
		server = new char[10];
		memcpy(server,"127.0.0.1",10);
		noServerEntered=true;
	}
	if (*server==0){
		delete server;
		server = new char[10];
		memcpy(server,"127.0.0.1",10);
		noServerEntered=true;
	}

	try{
		FILE* datei = fopen("wtris.conf","wb");
		if (datei!=NULL){
			fwrite(&namelen,sizeof(int),1,datei);
			fwrite(meinSpielerName,1,namelen,datei);
			if (noServerEntered){
				int a=0;
				fwrite(&a,sizeof(int),1,datei);
			}else{
				int serverLen = (int)strlen(server);
				fwrite(&serverLen,sizeof(int),1,datei);
				fwrite(server,1,serverLen,datei);
			}
			fclose(datei);
		}
	}catch(...){}


	WSADATA wsa;
	if (WSAStartup(MAKEWORD(2,0),&wsa)==0){
		ClientSocket=socket(AF_INET,SOCK_STREAM,0);
		if (ClientSocket!=INVALID_SOCKET){
			SOCKADDR_IN addr;
			memset(&addr,0,sizeof(SOCKADDR_IN));
			addr.sin_family=AF_INET;
			addr.sin_port=htons(PORT); // Tetris-Server-Port
			addr.sin_addr.s_addr=inet_addr(server); //Eingegebener Server

			bool ok = connect(ClientSocket,(SOCKADDR*)&addr,sizeof(SOCKADDR))!=SOCKET_ERROR;
			if (!ok && noServerEntered){
				startServerAtLocalhost();
				ok = connect(ClientSocket,(SOCKADDR*)&addr,sizeof(SOCKADDR))!=SOCKET_ERROR;
			}
			if (ok){
				//Buffer initialisieren
				char buffer[512];
				buffer[0]=0;

				//Jetzt senden ^^
				buffer[1] = OP_NICK;
				memcpy(buffer+2,meinSpielerName,namelen);
				SendBytes(ClientSocket,buffer,namelen+2);

				buffer[1] = OP_VERSION;
				buffer[2] = PROT_VERS_1;
				buffer[3] = PROT_VERS_2;
				buffer[4] = PROT_VERS_3;
				if (SendBytes(ClientSocket,buffer,5)){

					bool verbunden=true; //Netzwerkverbindung steht noch?


					//Mae der Texte ermitteln
					HDC hdc = GetDC(NULL);
					DrawTextA(hdc,"Nchster Stein:",-1,&nextStone_Text,DT_CALCRECT | DT_NOCLIP);
					int links = nextStone_Text.left+(nextStone_Text.right-nextStone_Text.left)/2-2*STONE_SIZE;
					if (links < 6){
						nextStone_Text.left-=links-6;
						nextStone_Text.right-=links-6;
					}
					{
						RECT tempRect;
						DrawTextA(hdc,"%s\r\n%d %s\r\n%d %s",-1,&tempRect,DT_CALCRECT | DT_NOCLIP);
						SpielerStatusHeight=tempRect.bottom-tempRect.top;
						DrawTextA(hdc,"Start",-1,&tempRect,DT_CALCRECT | DT_NOCLIP);
						ButtonTextHeight=tempRect.bottom-tempRect.top;
					}

					DeleteDC(hdc);

					//Knoepfe initialisieren ^^
					{
						int rechts = nextStone_Text.left + nextStone_Text.right-6;
						//links: 6
						//Abstand zwischen den Knoepfen: 6
						//unten: 20*STONE_SIZE+8
						//oben: wird berechnet, insgesamt 7 Button-Hoehen!
						int unten = 20*STONE_SIZE+8;
						int oben = unten - ((ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT)*7 + SPACE_BETWEEN_BUTTONS*(7-1));
						for (int i=0;i<13;i++){
							knoepfe[i].MousePressedItDown=false;
							knoepfe[i].DisplayItAsPressedDown=false;
							knoepfe[i].Coords.left=6;
							knoepfe[i].Coords.right=rechts;
						}

						memcpy(knoepfe[0].Caption,"Start",6);
						knoepfe[0].Coords.top=oben;
						knoepfe[0].Coords.bottom = knoepfe[0].Coords.top + ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT;
						knoepfe[0].pressedAction = ButtonStartPressed_Action;

						memcpy(knoepfe[1].Caption,"Pause",6);
						knoepfe[1].Coords.top=knoepfe[0].Coords.bottom+SPACE_BETWEEN_BUTTONS;
						knoepfe[1].Coords.bottom = knoepfe[1].Coords.top + ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT;
						knoepfe[1].pressedAction = ButtonPausePressed_Action;

						for (int i=2;i<11;i++){
							knoepfe[i].pressedAction = ButtonSpeedPressed_Action;
						}
						memcpy(knoepfe[2].Caption,"1",2);
						knoepfe[2].Coords.top=knoepfe[1].Coords.bottom+SPACE_BETWEEN_BUTTONS;
						knoepfe[2].Coords.bottom = knoepfe[2].Coords.top + ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT;
						knoepfe[2].Coords.right = (rechts-6-2*SPACE_BETWEEN_BUTTONS)/3 + 6;

						memcpy(knoepfe[3].Caption,"2",2);
						knoepfe[3].Coords.top=knoepfe[2].Coords.top;
						knoepfe[3].Coords.bottom = knoepfe[2].Coords.bottom;
						knoepfe[3].Coords.left = knoepfe[2].Coords.right + SPACE_BETWEEN_BUTTONS;
						knoepfe[3].Coords.right = (rechts-18)/3 + knoepfe[3].Coords.left;

						memcpy(knoepfe[4].Caption,"3",2);
						knoepfe[4].Coords.top=knoepfe[2].Coords.top;
						knoepfe[4].Coords.bottom = knoepfe[2].Coords.bottom;
						knoepfe[4].Coords.left = knoepfe[3].Coords.right + SPACE_BETWEEN_BUTTONS;

						for (int i=0;i<2;i++){
							for (int j=0;j<3;j++){
								knoepfe[5+3*i+j].Coords.top = knoepfe[2+3*i+j].Coords.bottom + SPACE_BETWEEN_BUTTONS;
								knoepfe[5+3*i+j].Coords.bottom = knoepfe[5+3*i+j].Coords.top + ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT;
								knoepfe[5+3*i+j].Coords.left = knoepfe[2+3*i+j].Coords.left;
								knoepfe[5+3*i+j].Coords.right = knoepfe[2+3*i+j].Coords.right;
							}
						}
						memcpy(knoepfe[5].Caption,"4",2);
						memcpy(knoepfe[6].Caption,"5",2);
						memcpy(knoepfe[7].Caption,"6",2);
						memcpy(knoepfe[8].Caption,"7",2);
						memcpy(knoepfe[9].Caption,"8",2);
						memcpy(knoepfe[10].Caption,"9",2);

						memcpy(knoepfe[11].Caption,"Modus",6);
						knoepfe[11].Coords.top = knoepfe[10].Coords.bottom + SPACE_BETWEEN_BUTTONS;
						knoepfe[11].Coords.bottom = knoepfe[11].Coords.top + ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT;
						knoepfe[11].pressedAction = ButtonModusPressed_Action;

						memcpy(knoepfe[12].Caption,"Reset",6);
						knoepfe[12].Coords.top = knoepfe[11].Coords.bottom + SPACE_BETWEEN_BUTTONS;
						knoepfe[12].Coords.bottom = knoepfe[12].Coords.top + ButtonTextHeight + 2*BUTTON_HEIGHT_FROM_TEXT;
						if (unten != knoepfe[12].Coords.bottom){
							printf("Warnung: Die Button-Positionen sind anders aus als geplant.\n");
						}
						knoepfe[12].pressedAction = ButtonResetPressed_Action;


						
					}

					//Jetzt Main-Window anlegen
					WNDCLASSEXA fensterklasse;
					fensterklasse.cbSize = sizeof(WNDCLASSEXA);
					fensterklasse.hCursor = LoadCursor(NULL,IDC_ARROW);
					fensterklasse.cbClsExtra = 0;
					fensterklasse.cbWndExtra = 0;
					fensterklasse.hIcon = LoadIcon(NULL,IDI_WINLOGO); //LoadIcon(hInstance,MAKEINTRESOURCE(IDI_ICON1));
					fensterklasse.hIconSm = NULL;
					fensterklasse.hInstance = hInstance;
					fensterklasse.lpfnWndProc = &MainDialogNachrichtenbehandlung;
					fensterklasse.lpszClassName = "wtrisWindow";
					fensterklasse.lpszMenuName = NULL;//MAKEINTRESOURCE(IDR_MENU1);
					fensterklasse.style = CS_HREDRAW | CS_VREDRAW;
					
					HBRUSH bgBrush = CreateSolidBrush(BG_COLOR);
					fensterklasse.hbrBackground = bgBrush;
					ATOM wndClassEx = RegisterClassExA(&fensterklasse);

					MainWindow = CreateWindowExA(0,"wtrisWindow",BEZEICHNUNG, WS_CAPTION|WS_SYSMENU|WS_VISIBLE, 0, 0, calculateWindowWidth(0),WINDOW_HEIGHT+6+SpielerStatusHeight,0,0,hInstance,0); //CreateWindowExA(WS_EX_APPWINDOW, fensterklasse.lpszClassName, BEZEICHNUNG, WS_CAPTION | WS_BORDER | WS_SYSMENU | WS_VISIBLE,0,0,160,320,HWND_DESKTOP,(HMENU)0,hInstance,NULL);
					if (!SetTimer(MainWindow, TIMER_ID, levelSpeeds[level-1], NULL))
						verbunden=false; //(dirty)-Trick zum verlassen der Schleife bei Timer-Fehler

					MSG msg = {0};

					HDC fenster = GetDC(MainWindow);
					TetrisBild = CreateCompatibleBitmap(fenster,STONE_SIZE*10+2,STONE_SIZE*20+2);
					Zeichenobjekt = CreateCompatibleDC(fenster);
					SelectObject(Zeichenobjekt, TetrisBild); //In Bitmap malen
					//Zeichenobjekt = GetDC(MainWindow);
					SelectObject(Zeichenobjekt,GetStockObject(DC_PEN));
					SelectObject(Zeichenobjekt,GetStockObject(DC_BRUSH));
					DeleteDC(fenster);

					SetDCBrushColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
					SetDCPenColor(Zeichenobjekt, 0x000000);
					Rectangle(Zeichenobjekt, 0, 0, 10*STONE_SIZE+2, 20*STONE_SIZE+2);
					InvalidateRect(MainWindow,NULL,FALSE);
					

					while(verbunden==true && msg.message!=WM_QUIT){
						if (PeekMessage(&msg,0,0,0,PM_REMOVE)){
							TranslateMessage(&msg); //Fenster-Nachrichten verarbeiten
							DispatchMessage(&msg);
							continue;
						}

						//Netzwerk-Nachrichten lesen
						timeval WarteOption = {0, 0};
						fd_set sockets;
						sockets.fd_count=1;
						sockets.fd_array[0]=ClientSocket;
						char* rBuffer = NULL;
						int bufferlen = 0;
						while (select(0,&sockets, NULL, NULL, &WarteOption)>0){
							//Lesen ^^
							char inBuffer[4096];
							int laenge = recv(ClientSocket, inBuffer, 4096, 0);
							if (laenge <= 0){ //Verbindung fehlgeschlagen
								verbunden=false;
								break;
							}
							char* temp = rBuffer;
							rBuffer = new char[laenge+bufferlen];
							if (temp!=NULL){
								memcpy(rBuffer,temp,bufferlen);
							}
							memcpy(rBuffer+bufferlen,inBuffer,laenge);
							if (temp!=NULL){
								delete temp;
							}
							bufferlen +=laenge;
						}
						if (rBuffer!=NULL){
							//rBuffer auswerten!!!
							char* buf = rBuffer;
							while(bufferlen > 3){
								buf+=3;
								int befehlslaenge=(int)((unsigned char)(*buf));
								buf++;
								bufferlen-=4;
								if (bufferlen < befehlslaenge)
									break;
								switch ((int)((unsigned int)(*(buf+1)))){
									//Befehl lesen (buf[0]...buf[befehlslaenge-1])

									case OP_NICK:
										{
											char* nick = new char[befehlslaenge-1];
											memcpy(nick,buf+2,befehlslaenge-2);
											nick[befehlslaenge-2]=0;
											printf("Spieler %d hat den Nick \"%s\".\n",buf[0],nick);
											int position = getInternGegnerNumber(buf[0]);
											if (position<0||position>=anzahlGegner){
												printf("Warnung: Spieler %d nicht gefunden, Nickname wird verworfen.\n",buf[0]);
												delete nick;
												break;
											}
											memcpy(gegner[position].name,nick,befehlslaenge-2<16?befehlslaenge-2:16);
											if (befehlslaenge-2<16)
												gegner[position].name[befehlslaenge-2]=0;
											else
												gegner[position].name[15]=0;
											//Statistik-Bereich neu zeichnen ^^
											RECT statistik = {0,14+20*STONE_SIZE,calculateWindowWidth(anzahlGegner),14+20*STONE_SIZE+SpielerStatusHeight};
											InvalidateRect(MainWindow,&statistik,FALSE);
											delete nick;
											break;
										}
									case OP_PLAY:
										if (buf[0]==0){
											printf("Der Server hat ein neues Spiel gestartet.\n");
										}else{
											if (buf[0]==meineSpielerID){
												printf("Ich habe ein neues Spiel gestartet.\n");
											}else{
												int position = getInternGegnerNumber(buf[0]);
												if (0<=position && position<anzahlGegner)
													printf("%s hat ein neues Spiel gestartet.\n",gegner[position].name);
												else{
													printf("Spieler %d hat ein neues Spiel gestartet.\n",buf[0]);
												}
											}
										}
										//Neues Spiel starten
										spielLaeuft=true;
										addlines=0;
										addfun=0;
										lines=0;
										loseAnimationStatus = -1;
										if (loseAnimationActive){
											KillTimer(MainWindow,LOSE_TIMER_ID);
											loseAnimationActive=false;
										}

										for (int i=0;i<anzahlGegner;i++){
											gegner[i].linien=0;
											gegner[i].loseAnimationStatus = -1;
											if (gegner[i].loseAnimationActive){
												KillTimer(MainWindow,LOSE_TIMER_ID+1+i);
												gegner[i].loseAnimationActive=false;
											}
										}
										cleanSpielfeld();

										SetDCBrushColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
										SetDCPenColor(Zeichenobjekt, SPIELFELD_BG_COLOR);
										Rectangle(Zeichenobjekt, 1, 1, 10*STONE_SIZE+1, 20*STONE_SIZE+1);
										InvalidateRect(MainWindow,NULL,FALSE);

										buffer[1]=OP_CLEAR;
										SendBytes(ClientSocket,buffer,2);

										if (fun){
											//Spielfeld vorab fuellen
											//Erst am Rand,
											int redraw[73];
											createFunModeSides(redraw);
											redrawLocalAndSendToServer(MainWindow,Zeichenobjekt,redraw,ClientSocket);

											//Dann in der Mitte
											createFunModeMiddle(redraw);
											redrawLocalAndSendToServer(MainWindow,Zeichenobjekt,redraw,ClientSocket);
										}

										naechsterStein.steinArt=(STEIN)(rand()%7);
										int redraw[25];
										legeNeuenSteinAn(ClientSocket,MainWindow,redraw,Zeichenobjekt);
										//createNewStone(&dieserStein,&naechsterStein,redraw);
										//redrawLocalAndSendToServer(Zeichenobjekt,redraw,ClientSocket);
										//SetTimer(MainWindow, TIMER_ID, levelSpeeds[level-1], NULL);

										break;
									case OP_FALL:
										{
											int position = getInternGegnerNumber(buf[0]);
											if (position<0||position>=anzahlGegner)
												break;
											for (int i=2;i<befehlslaenge;i++){
												BitBlt(gegner[position].Zeichenobjekt,1,STONE_SIZE+1,10*STONE_SIZE,STONE_SIZE*buf[i],gegner[position].Zeichenobjekt,1,1,SRCCOPY);
												SetDCBrushColor(gegner[position].Zeichenobjekt, SPIELFELD_BG_COLOR);
												SetDCPenColor(gegner[position].Zeichenobjekt, SPIELFELD_BG_COLOR);
												Rectangle(gegner[position].Zeichenobjekt, 1, 1, 10*STONE_SIZE+1, STONE_SIZE+1);
												gegner[position].linien++;
											}
											//Gegner-Bereich neu zeichnen
											RECT invalidateRect = {nextStone_Text.right+nextStone_Text.left-6+10*STONE_SIZE+14+(STONE_SIZE*10+6)*position,6,0,14+20*STONE_SIZE+SpielerStatusHeight};
											invalidateRect.right=invalidateRect.left+STONE_SIZE*10+2;
											InvalidateRect(MainWindow,&invalidateRect,FALSE);
										}
										break;
									case OP_DRAW:
										{
											int position = getInternGegnerNumber(buf[0]);
											if (position<0||position>=anzahlGegner)
												break;
											for (int i=2;i<befehlslaenge-2;i+=3){
												DrawBlock(gegner[position].Zeichenobjekt,buf[i],buf[i+1],buf[i+2]);
											}
											//Gegner-Bereich neu zeichnen
											RECT invalidateRect = {nextStone_Text.right+nextStone_Text.left-6+10*STONE_SIZE+14+(STONE_SIZE*10+6)*position,6,0,STONE_SIZE*20+8};
											invalidateRect.right=invalidateRect.left+STONE_SIZE*10+2;
											InvalidateRect(MainWindow,&invalidateRect,FALSE);
										}
										break;
									case OP_LOST:
										{
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner){
												printf("%s hat verloren.\n",gegner[position].name);
												//Lose-Animation starten
												gegner[position].loseAnimationStatus=-1;
												SetTimer(MainWindow, LOSE_TIMER_ID+position+1, LOSE_ANIMATION_SPEED, NULL);
												gegner[position].loseAnimationActive=true;
											}else
												printf("Spieler %d hat verloren.\n",buf[0]);
										}
										break;
									case OP_GONE:
										{
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner)
												printf("%s hat das Spiel verlassen.\n",gegner[position].name);
											else{
												printf("Spieler %d hat das Spiel verlassen.\nWarnung: Spieler %d konnte nicht entfernt werden, weil er nicht existiert.\n",buf[0],buf[0]);
												break;
											}
											ReleaseDC(MainWindow,gegner[position].Zeichenobjekt);
											DeleteObject(gegner[position].TetrisBild);
											for (int i=position; i<anzahlGegner-1;i++){
												gegner[i] = gegner[i+1];
											}
											anzahlGegner--;
											changeWindowWidth(MainWindow,calculateWindowWidth(anzahlGegner));
										}
										break;
									case OP_CLEAR:
										{
											int position = getInternGegnerNumber(buf[0]);
											if (position<0||position>=anzahlGegner)
												break;
											SetDCBrushColor(gegner[position].Zeichenobjekt, SPIELFELD_BG_COLOR);
											SetDCPenColor(gegner[position].Zeichenobjekt, 0x000000);
											Rectangle(gegner[position].Zeichenobjekt, 0, 0, 10*STONE_SIZE+2, 20*STONE_SIZE+2);
										}
										break;
									case OP_NEW:
										{
											printf("Spieler %d ist dem Spiel beigetreten.\n",buf[0]);
											//buf[2+3]: unsigned short: anzahl der gewonnenen spiele
											Gegenspieler* temp = gegner;
											gegner = new Gegenspieler[anzahlGegner+1];
											if (temp!=NULL){
												memcpy(gegner,temp,sizeof(Gegenspieler)*anzahlGegner);
											}
											gegner[anzahlGegner].gewonnen=0;
											if (befehlslaenge>=4){
												((char*)&(gegner[anzahlGegner].gewonnen))[1] = buf[2];
												((char*)&(gegner[anzahlGegner].gewonnen))[0] = buf[3];
											}
											gegner[anzahlGegner].id=buf[0];
											gegner[anzahlGegner].linien=0;
											gegner[anzahlGegner].name[0]=0;

											HDC fenster = GetDC(MainWindow);
											gegner[anzahlGegner].TetrisBild = CreateCompatibleBitmap(fenster,STONE_SIZE*10+2,STONE_SIZE*20+2);
											gegner[anzahlGegner].Zeichenobjekt = CreateCompatibleDC(fenster);
											SelectObject(gegner[anzahlGegner].Zeichenobjekt, gegner[anzahlGegner].TetrisBild); //In Bitmap malen
											SelectObject(gegner[anzahlGegner].Zeichenobjekt,GetStockObject(DC_PEN));
											SelectObject(gegner[anzahlGegner].Zeichenobjekt,GetStockObject(DC_BRUSH));
											DeleteDC(fenster);

											SetDCBrushColor(gegner[anzahlGegner].Zeichenobjekt, SPIELFELD_BG_COLOR);
											SetDCPenColor(gegner[anzahlGegner].Zeichenobjekt, 0x000000);
											Rectangle(gegner[anzahlGegner].Zeichenobjekt, 0, 0, 10*STONE_SIZE+2, 20*STONE_SIZE+2);

											gegner[anzahlGegner].loseAnimationStatus = -1;
											gegner[anzahlGegner].loseAnimationActive = false;

											anzahlGegner++;
											if (temp!=NULL){
												delete temp;
											}
											changeWindowWidth(MainWindow,calculateWindowWidth(anzahlGegner));
										}
										break;
									case OP_LINES:
										{
											//Linien bekommen -_-
											int lines = buf[2];
											if (fun){
												if (lines==1){
													//Fun-Block ^^
													int position = getInternGegnerNumber(buf[0]);
													if (0<=position && position<anzahlGegner)
														printf("Ich habe einen fun-Block von %s bekommen.\n",gegner[position].name);
													else
														printf("Ich habe einen fun-Block von Spieler %d bekommen.\n",buf[0]);
													addfun++;
													break;
												}
												lines=lines*2-3;
											}
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner)
												printf("Ich habe %d Linien von %s bekommen... -_-\n",lines,gegner[position].name);
											else
												printf("Ich habe %d Linien von Spieler %d bekommen... -_-\n",lines,buf[0]);
											addlines+=lines;
											break;
										}
									case OP_GROW:
										{
											int position = getInternGegnerNumber(buf[0]);
											if (position<0||position>=anzahlGegner)
												break;
											BitBlt(gegner[position].Zeichenobjekt,1,1,10*STONE_SIZE,STONE_SIZE*19,gegner[position].Zeichenobjekt,1,STONE_SIZE+1,SRCCOPY);
											SetDCBrushColor(gegner[position].Zeichenobjekt, SPIELFELD_BG_COLOR);
											SetDCPenColor(gegner[position].Zeichenobjekt, SPIELFELD_BG_COLOR);
											Rectangle(gegner[position].Zeichenobjekt, 1, 19*STONE_SIZE+1, 10*STONE_SIZE+1, 20*STONE_SIZE+1);
											//Gegner-Bereich neu zeichnen
											RECT invalidateRect = {nextStone_Text.right+nextStone_Text.left-6+10*STONE_SIZE+14+(STONE_SIZE*10+6)*position,6,0,STONE_SIZE*20+8};
											invalidateRect.right=invalidateRect.left+STONE_SIZE*10+2;
											InvalidateRect(MainWindow,&invalidateRect,FALSE);
										}
										break;
									case OP_MODE:
										//normal, fun, unknown...
										if (buf[0]!=0){
											if (buf[0]==meineSpielerID){
												printf("Ich habe den Modus \"%s%s\" gesetzt.\n",buf[2]%2==MODUS_NORMAL?"normal":(buf[2]%2==MODUS_FUN?"fun":"unknown"),((buf[2]/2)%2==1?"/speed":""));
											}else{
												int position = getInternGegnerNumber(buf[0]);
												if (0<=position && position<anzahlGegner)
													printf("%s hat den Modus \"%s\" gesetzt.\n",gegner[position].name,buf[2]%2==MODUS_NORMAL?"normal":(buf[2]%2==MODUS_FUN?"fun":"unknown"));
												else
													printf("Spieler %d hat den Modus \"%s\" gesetzt.\n",buf[0],buf[2]%2==MODUS_NORMAL?"normal":(buf[2]%2==MODUS_FUN?"fun":"unknown"));
											}
										}else{
											printf("Der Server hat den Modus \"%s\" gesetzt.\n",buf[2]%2==MODUS_NORMAL?"normal":(buf[2]%2==MODUS_FUN?"fun":"unknown"));
										}
										if (buf[2]%2==MODUS_FUN)
											fun=true;
										else
											fun=false;
										if ((buf[2]/2)%2==MODUS_NORMAL)
											speed_mode = false;
										else
											speed_mode = true;
										InvalidateRect(MainWindow,NULL,false);
										break;
									case OP_LEVEL:
										//Speed aendern
										level = buf[2];
										if (level < 1)
											level=1;
										else if (level > 9)
											level = 9;
										if (buf[0]!=0){
											if (buf[0]==meineSpielerID){
												printf("Ich habe das Tempo auf Stufe %d gesetzt.\n",level);
											}else{
												int position = getInternGegnerNumber(buf[0]);
												if (0<=position && position<anzahlGegner)
													printf("%s hat das Tempo auf Stufe %d gesetzt.\n",gegner[position].name,level);
												else
													printf("Spieler %d hat das Tempo auf Stufe %d gesetzt.\n",buf[0],level);
											}
										}else{
											printf("Der Server hat das Tempo auf Stufe %d gesetzt.\n",level);
										}
										KillTimer(MainWindow, TIMER_ID);
										SetTimer(MainWindow, TIMER_ID, levelSpeeds[level-1], NULL);
										if (buf[0]==meineSpielerID){
											SendMessage(MainWindow,WM_TIMER,(WPARAM)TIMER_ID,0);
										}
										InvalidateRect(MainWindow,NULL,false);
										break;
										/* //OP_BOT und OP_KILL werden nur an den Server geschickt.
									case OP_BOT:
										break;
									case OP_KILL:
										break;
										*/
									case OP_PAUSE:
										//Spiel anhalten
										if (buf[0]==meineSpielerID){
											printf("Ich habe das Spiel angehalten.\n");
										}else{
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner)
												printf("%s hat das Spiel angehalten.\n",gegner[position].name);
											else
												printf("Spieler %d hat das Spiel angehalten.\n",buf[0]);
										}
										spielPausiert=true;
										if (spielLaeuft){
											spielLaeuft=false;
											pause=true;
											KillTimer(MainWindow, TIMER_ID);
										}
										memcpy(knoepfe[1].Caption,"Weiter",7);
										InvalidateRect(MainWindow,&(knoepfe[1].Coords),FALSE);
										break;
									case OP_CONT:
										//Spiel fortsetzen
										if (buf[0]==meineSpielerID){
											printf("Ich habe das Spiel fortgesetzt.\n");
										}else{
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner)
												printf("%s hat das Spiel fortgesetzt.\n",gegner[position].name);
											else
												printf("Spieler %d hat das Spiel fortgesetzt.\n",buf[0]);
										}
										spielPausiert=false;
										if (pause){
											pause=false;
											spielLaeuft=true;
											SetTimer(MainWindow, TIMER_ID, levelSpeeds[level-1], NULL);
											SendMessage(MainWindow,WM_TIMER,(WPARAM)TIMER_ID,0); //Um Missbrauch vorzubeugen
										}
										memcpy(knoepfe[1].Caption,"Pause",6);
										InvalidateRect(MainWindow,&(knoepfe[1].Coords),FALSE);
										break;
									case OP_VERSION:
										//Protokoll-Version erhalten
										if (buf[2]==PROT_VERS_1 && buf[3] >= PROT_VERS_2 || buf[2] > PROT_VERS_1)
											if (buf[2]==PROT_VERS_1 && buf[3]==PROT_VERS_2)
												server_prot_revision = buf[4];
											else
												server_prot_revision = (PROT_VERS_3>2?PROT_VERS_3:2);
										break;
									case OP_BADVERS:
										//Schade -_-
										printf("Der Server meldete einen Versionskonflikt der Netzwerkprotokolle.\n");
										break;
									case OP_MSG:
										{
											//Erhalte Chat-Nachricht ^^
											char* nachricht = new char[befehlslaenge-1];
											memcpy(nachricht,buf+2,befehlslaenge-2);
											nachricht[befehlslaenge-2]=0;
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner)
												printf("%s schrieb: %s\n",gegner[position].name,nachricht);
											else
												printf("Spieler %d schrieb: %s\n",buf[0],nachricht);
											delete nachricht;
											break;
										}
									case OP_YOUARE:
										//Erhalte ID ^^
										meineSpielerID = buf[0];
										printf("Ich bin Spieler %d.\n",buf[0]);
										break;
									case OP_LINESTO:
										//a schickt an b ^^
										if (buf[0] == meineSpielerID){
											int position = getInternGegnerNumber(buf[3]);
											if (buf[2]==1 && fun){
												if (0<=position && position<anzahlGegner)
													printf("Ich habe einen fun-Block an %s geschickt.\n",gegner[position].name);
												else
													printf("Ich habe einen fun-Block an Spieler %d geschickt.\n",buf[3]);
											}else{
												if (0<=position && position<anzahlGegner)
													printf("Ich habe %d Zeilen an %s geschickt.\n",fun?(buf[2]*2-3):buf[2],gegner[position].name);
												else
													printf("Ich habe %d Zeilen an Spieler %d geschickt.\n",fun?(buf[2]*2-3):buf[2],buf[3]);
											}
										}else{
											//OP_LINESTO bei nicht-Betroffensein nicht ausgeben.
											/*
											if (buf[2]==1 && fun){
												printf("Spieler %d schickt einen fun-Block an %d.\n",buf[0],buf[3]);
											}else{
												printf("Spieler %d schickt %d Reihen an %d.\n",buf[0],fun?(buf[2]*2-3):buf[2],buf[3]);
											}
											*/
										}
										break;
									case OP_WON:
										if (buf[0]==meineSpielerID){
											//Juhu, gewonnen ^^
											printf("Ich habe gewonnen! ^_^\n",buf[0]);
											//TODO: evtl. WIN-Animation bauen
											gewonnen++;
											//Statistik neu zeichnen
											RECT statistik = {0,14+20*STONE_SIZE,calculateWindowWidth(anzahlGegner),14+20*STONE_SIZE+SpielerStatusHeight};
											InvalidateRect(MainWindow,&statistik,FALSE);
										}else{
											int position = getInternGegnerNumber(buf[0]);
											if (0<=position && position<anzahlGegner){
												printf("%s hat gewonnen.\n",gegner[position].name);
												gegner[position].gewonnen++;
												RECT statistik = {0,14+20*STONE_SIZE,calculateWindowWidth(anzahlGegner),14+20*STONE_SIZE+SpielerStatusHeight};
												InvalidateRect(MainWindow,&statistik,FALSE);
											}else
												printf("Spieler %d hat gewonnen.\n",buf[0]);
										}
										break;
									case OP_ZERO:
										//Spielstaende zuruecksetzen
										if (buf[0]!=0){
											if (buf[0]==meineSpielerID){
												printf("Ich habe die Spielstaende zurueckgesetzt.\n");
											}else{
												int position = getInternGegnerNumber(buf[0]);
												if (0<=position && position<anzahlGegner)
													printf("%s hat die Spielstaende zurueckgesetzt.\n",gegner[position].name);
												else
													printf("Spieler %d hat die Spielstaende zurueckgesetzt.\n",buf[0]);
											}
										}else{
											printf("Der Server hat die Spielstaende zurueckgesetzt.\n");
										}
										{
											gewonnen=0;
											lines=0;
											for (int i=0;i<anzahlGegner;i++){
												gegner[i].gewonnen=0;
												gegner[i].linien=0;
											}
											RECT statistik = {0,14+20*STONE_SIZE,calculateWindowWidth(anzahlGegner),14+20*STONE_SIZE+SpielerStatusHeight};
											InvalidateRect(MainWindow,&statistik,FALSE);
										}
										break;
									default:
										//unknown command
										printf("Unbekannte Nachricht erhalten: %d\n",buf[1]);
										break;

								}
								buf+=befehlslaenge;
								bufferlen-=befehlslaenge;
								befehlslaenge=0;
							}
							//Wenn der Server das Spiel startet: spielLaeuft=true setzen

							delete rBuffer;
						}


						Sleep(1); //etwas Rechenzeit abgeben ^^
					}
					KillTimer(MainWindow, TIMER_ID);
					ReleaseDC(MainWindow,Zeichenobjekt);
					DeleteObject(TetrisBild);

				}
			}
		}
		closesocket(ClientSocket);
	}
	//Speicher freigeben
	delete server;
	server=NULL;
	if (anzahlGegner>0)
		if (gegner!=NULL){
			for (int i=0;i<anzahlGegner;i++){
				ReleaseDC(MainWindow,gegner[i].Zeichenobjekt);
				DeleteObject(gegner[i].TetrisBild);
			}
			delete gegner;
		}
	return 0;
}

void toggleMode(){
	char buffer[3];
	buffer[0]=0;
	buffer[1]=OP_MODE;
	buffer[2]=fun?MODUS_NORMAL:MODUS_FUN;
	buffer[2]+=(server_prot_revision>=2?((fun && !speed_mode || !fun && speed_mode)?MODUS_SPEED:0):0);
	SendBytes(ClientSocket,buffer,3);
}